//**************************************************************************
//*                     This file is part of the                           *
//*                MMC - Mpxplay Multimedia Commander                      *
//*                   The source code of MMC is                            *
//*        (C) copyright 1998-2020 by PDSoft (Attila Padar)                *
//*                http://mpxplay.sourceforge.net                          *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: (bottom) video control bar

//#define MPXPLAY_USE_DEBUGF 1
#define DISPQT_DEBUG_OUTPUT stdout

#include <QtGui>
#include <QtWidgets>
#include <QtWinExtras/QtWinExtras>
#include "disp_qt.h"
#include "moc_video_ctrlbar.h"
#include "moc_seekbar.h"
#include "moc_mainwindow.h"
#include "moc_video_qt.h"

#define DISPQT_VIDEOCTRLBAR_SIZE_Y_AERO  80 // Win7 Aero window cannot be smaller than this!
#define DISPQT_VIDEOCTRLBAR_SIZE_Y_BLACK 70 // simple black or Win10 transparent window (headerless)

#define DISPQT_VIDEOCTRLBAR_OPACITY_MIN             0 //
#define DISPQT_VIDEOCTRLBAR_OPACITY_MAX           100 //
#define DISPQT_VIDEOCTRLBAR_FADEOUT_TIMERINTERVAL  18 // fadeout timer interval
#define DISPQT_VIDEOCTRLBAR_FADEOUT_OPACITYSTEP    10 // fadeout opacity steps -> 20*10 = 200ms to fade out

extern "C" {
 extern struct mainvars mvps;
 extern unsigned int refdisp, timemode;
 extern keyconfig kb[];
}

DispQtVideoControlbar::DispQtVideoControlbar(MainWindow *main_window, QWidget *parent) : DispQtDialogElemWindow(main_window, parent, Qt::CoverWindow | Qt::WindowDoesNotAcceptFocus, DISPQT_DIALOG_WINDOWTYPE_VIDEOCTRLBAR)
{
	this->main_window = main_window;
	this->is_show_enabled = false;
	this->is_win7aero_window = false;
	this->opacity_curr = DISPQT_VIDEOCTRLBAR_OPACITY_MAX;
	this->timer_fadeout.setSingleShot(false);

	this->setFocusPolicy(Qt::NoFocus);

	this->videoctrlbar_config_style_apply(true);

	this->seek_slider = new DispQtMainwinSeekslider(main_window, this, true);

	this->number = new QLabel;
	this->number->setContentsMargins(0,0,0,0);
	QFont font = this->number->font();
	font.setPointSize(12);
	this->number->setFont(font);

	optionToolBar = new QToolBar(tr("VidOpt"), this);
	optionToolBar->setParent(this);
	optionToolBar->setContentsMargins(0,0,0,0);
	optionToolBar->addAction(main_window->action_view_videofs);
	optionToolBar->addAction(main_window->action_list_replay);
	optionToolBar->addAction(main_window->action_play_highlight);
	optionToolBar->addAction(main_window->action_view_videowall);

	playToolBar = new QToolBar(tr("VidPlay"), this);
	playToolBar->setParent(this);
	playToolBar->setContentsMargins(0,0,0,0);
	playToolBar->addAction(main_window->action_play_skipback);
	playToolBar->addAction(main_window->action_play_seekback);
	playToolBar->addAction(main_window->action_play_start);
	playToolBar->addAction(main_window->action_play_stop);
	playToolBar->addAction(main_window->action_play_seekforward);
	playToolBar->addAction(main_window->action_play_skipforward);

	QToolButton *button_rewind = dynamic_cast<QToolButton*>(playToolBar->widgetForAction(main_window->action_play_seekback));
	button_rewind->setAutoRepeat(true);
	button_rewind->setAutoRepeatDelay(100);
	button_rewind->setAutoRepeatInterval(1000/30);
	QToolButton* button_forward = dynamic_cast<QToolButton*>(playToolBar->widgetForAction(main_window->action_play_seekforward));
	button_forward->setAutoRepeat(true);
	button_forward->setAutoRepeatDelay(100);
	button_forward->setAutoRepeatInterval(1000/30);

	QVBoxLayout *layout_ctrlbar = new QVBoxLayout;
	layout_ctrlbar->setMargin(0);
	layout_ctrlbar->setSpacing(0);
	layout_ctrlbar->addWidget(this->seek_slider, 1);

	QWidget *widget_play = new QWidget;
	QHBoxLayout *layout_play = new QHBoxLayout;
	layout_play->setMargin(0);
	layout_play->setSpacing(0);
	layout_play->addWidget(optionToolBar, 1, Qt::AlignLeft | Qt::AlignVCenter);
	layout_play->addWidget(playToolBar, 1, Qt::AlignCenter);
	layout_play->addWidget(this->number, 1, Qt::AlignRight | Qt::AlignVCenter);
	widget_play->setLayout(layout_play);
	layout_ctrlbar->addWidget(widget_play, 0);

	this->setLayout(layout_ctrlbar);

	connect(this, SIGNAL(signal_videoctrlbar_config_style_apply(bool)), this, SLOT(videoctrlbar_config_style_apply(bool)));
	connect(this, SIGNAL(signal_videoctrlbar_set_show(bool)), this, SLOT(videoctrlbar_set_show(bool)));
	connect(this, SIGNAL(signal_videoctrlbar_set_timepos(int)), this, SLOT(videoctrlbar_set_timepos(int)));
	connect(this, SIGNAL(signal_videoctrlbar_option_timemode_change(unsigned int)), this, SLOT(videoctrlbar_option_timemode_change(unsigned int)));
	connect(this, SIGNAL(signal_videoctrlbar_seekbar_send_seekpreview(void *)), this, SLOT(videoctrlbar_seekbar_send_seekpreview(void *)));
	connect(&timer_fadeout, SIGNAL(timeout()), this, SLOT(videoctrlbar_fadeout_window_timerfunc()));
	this->hide();
}

DispQtVideoControlbar::~DispQtVideoControlbar()
{
	if(this->seek_slider)
	{
		DispQtMainwinSeekslider *mss = this->seek_slider;
		this->seek_slider = NULL;
		delete mss;
	}
}

bool DispQtVideoControlbar::event(QEvent *event)
{
	bool cursor_hide = false;
	switch(event->type()){
		case QEvent::Enter: this->videoctrlbar_show_window(); break;
		case QEvent::Leave: if(!this->is_win7aero_window)
		                    	this->videoctrlbar_fadeout_window_start();
		                    cursor_hide = true;
		                    break;
		case QEvent::Move:  this->videoctrlbar_set_size_and_pos(); break;
		default: goto err_out_event;
	}
	if(this->main_window->mainwin_is_on_video())
		emit this->main_window->signal_video_mouse_cursor_show(cursor_hide);

err_out_event:
	return QDialog::event(event);
}

void DispQtVideoControlbar::keyPressEvent(QKeyEvent *event)
{
	this->main_window->handle_keypressevent(event);
}

void DispQtVideoControlbar::wheelEvent(QWheelEvent *event)
{
	this->main_window->mainwin_handle_keypresscode(kb[(event->delta() < 0)? 0 : 2].c);  // KeyRewind1 / KeyForward1
	event->accept();
}

void DispQtVideoControlbar::mousePressEvent(QMouseEvent *event)
{
	Qt::MouseButton ebutton = event->button();
	if((ebutton & Qt::MidButton) || (ebutton & Qt::RightButton))
		this->main_window->mainwin_handle_mousepress_skip_or_seek_event(event);
	else
		QDialog::mousePressEvent(event);
}

void DispQtVideoControlbar::mouseReleaseEvent(QMouseEvent *event)
{
	Qt::MouseButton ebutton = event->button();
	if((ebutton & Qt::MidButton) || (ebutton & Qt::RightButton))
		this->main_window->mainwin_handle_mousepress_skip_or_seek_event(event);  // to handle press or click
	else
		QDialog::mouseReleaseEvent(event);
}

void DispQtVideoControlbar::videoctrlbar_set_size_and_pos(void)
{
	int max_x = QApplication::desktop()->screenGeometry().width();
	int max_y = QApplication::desktop()->screenGeometry().height();
	int winsize_x = max_x * this->main_window->gui_config->videocontrolbar_size_percent / 100;
	int winsize_y = (this->is_win7aero_window)? DISPQT_VIDEOCTRLBAR_SIZE_Y_AERO : DISPQT_VIDEOCTRLBAR_SIZE_Y_BLACK;

	this->setFixedSize(winsize_x, winsize_y);
	this->setContentsMargins(0, 0, 10, (this->is_win7aero_window)? 20 : 10);
	this->move((max_x - winsize_x) / 2, max_y - winsize_y - ((this->is_win7aero_window)? 5 : 0));
}

void DispQtVideoControlbar::videoctrlbar_config_style_apply(bool initial)
{
	bool lucent_enable = this->main_window->mainwin_is_translucent_enabled();
	bool glass_ctrlbar = (lucent_enable && !funcbit_test(this->main_window->gui_config->video_control, DISPQT_CONFIG_VIDEOCTRL_CONTROLBAR_BLACK));

	this->is_win7aero_window = (glass_ctrlbar && (this->main_window->ms_windows_version == MPXPLAY_MSWIN_VERSIONID_WIN7))? true : false;

	this->videoctrlbar_set_size_and_pos();

	if(glass_ctrlbar){
#ifdef Q_OS_WIN
		if(this->main_window->ms_windows_version > MPXPLAY_MSWIN_VERSIONID_WIN7)
			mpxplay_dispqt_dialog_elems_setwindowcomposition_win10((HWND)this->window()->winId(), DISPQT_DIALOG_HEXACOLOR_BLACK);
#endif
		this->setWindowFlags(Qt::CoverWindow | Qt::WindowDoesNotAcceptFocus);
		QtWin::extendFrameIntoClientArea(this, -1, -1, -1, -1);
		setAttribute(Qt::WA_TranslucentBackground, true);
		setAttribute(Qt::WA_NoSystemBackground, false);
		setStyleSheet("QDialog { background: transparent; } QLabel { color:white; }");
	} else {
		if(!initial)
		{
#ifdef Q_OS_WIN
			if(this->main_window->ms_windows_version > MPXPLAY_MSWIN_VERSIONID_WIN7)
				mpxplay_dispqt_dialog_elems_reset_windowcomposition_win10((HWND)this->window()->winId());
#endif
			QtWin::resetExtendedFrame(this);
		}
		this->setWindowFlags(Qt::Dialog | Qt::FramelessWindowHint | Qt::WindowDoesNotAcceptFocus);
		setAttribute(Qt::WA_TranslucentBackground, true);
		setAttribute(Qt::WA_NoSystemBackground, false);
		setStyleSheet("QDialog { background: rgba(0,0,0,20); } QLabel { color:white; }");
	}
	this->setWindowTitle(" ");
	if(!initial)
		this->seek_slider->mainwin_seekslider_config_style_apply(lucent_enable);
}

void DispQtVideoControlbar::videoctrlbar_set_show(bool enable)
{
	if(!enable && this->is_show_enabled)
		this->videoctrlbar_hide_window();
	this->is_show_enabled = enable;
}

bool DispQtVideoControlbar::videoctrlbar_check_show(int x, int y)
{
	if(!this->is_show_enabled || !funcbit_test(this->main_window->gui_config->video_control, DISPQT_CONFIG_VIDEOCTRL_ENABLE_CONTROLBAR))
		return false;
	int pos_x = this->pos().x(), pos_y = this->pos().y(), size_x = this->width(), size_y = this->height();
	if((x >= pos_x) && (x < (pos_x + size_x)) && (y >= pos_y) && (y < (pos_y + size_y))) {
		this->videoctrlbar_show_window();
		return true;
	} else
		this->videoctrlbar_fadeout_window_start();
	return false;
}

void DispQtVideoControlbar::videoctrlbar_show_window(void)
{
	if(!this->is_show_enabled && !funcbit_test(this->main_window->gui_config->video_control, DISPQT_CONFIG_VIDEOCTRL_ENABLE_CONTROLBAR))
		return;
	this->videoctrlbar_set_size_and_pos();
	this->opacity_curr = DISPQT_VIDEOCTRLBAR_OPACITY_MAX;
	this->setWindowOpacity((qreal)this->opacity_curr / (qreal)DISPQT_VIDEOCTRLBAR_OPACITY_MAX);
	this->timer_fadeout.stop();
	this->show();
}

void DispQtVideoControlbar::videoctrlbar_hide_window(void)
{
	this->seek_slider->seekslider_preview_hide();
	this->timer_fadeout.stop();
	this->hide();
	this->main_window->raise(); // to drop always on top to parent window
}

void DispQtVideoControlbar::videoctrlbar_fadeout_window_start(void)
{
	if(this->isHidden() || this->timer_fadeout.isActive())
		return;
	if(funcbit_test(this->main_window->gui_config->video_control, DISPQT_CONFIG_VIDEOCTRL_CONTROLBAR_FADEOUT))
	{
		this->opacity_curr = DISPQT_VIDEOCTRLBAR_OPACITY_MAX;
		this->timer_fadeout.start(DISPQT_VIDEOCTRLBAR_FADEOUT_TIMERINTERVAL);
	}
	else
	{
		this->videoctrlbar_hide_window();
	}
}

void DispQtVideoControlbar::videoctrlbar_fadeout_window_timerfunc(void)
{
	if(this->opacity_curr > DISPQT_VIDEOCTRLBAR_OPACITY_MIN)
	{
		this->setWindowOpacity((qreal)this->opacity_curr / (qreal)DISPQT_VIDEOCTRLBAR_OPACITY_MAX);
		this->opacity_curr -= DISPQT_VIDEOCTRLBAR_FADEOUT_OPACITYSTEP;
	}
	else
	{
		this->videoctrlbar_hide_window();
	}
}

void DispQtVideoControlbar::videoctrlbar_set_currfileinfos(struct dispqt_currfileinfo_s *cur)
{
	this->time_len_ms = cur->currtimelen;
	this->seek_slider->mainwin_seekslider_set_timelen(cur->currtimelen);
	this->setWindowTitle(QString::fromUtf8(cur->curr_title));
	mpxplay_debugf(DISPQT_DEBUG_OUTPUT,"title: %s", cur->curr_title);
}

void DispQtVideoControlbar::videoctrlbar_set_timepos(int new_timepos_ms)
{
	char sout[128];
	emit this->seek_slider->mainwin_seekslider_set_timepos(new_timepos_ms);
	if(this->time_len_ms < 3600000) {
		snprintf(sout, sizeof(sout), "%d:%2.2d / %d:%2.2d", (new_timepos_ms / 60000), ((new_timepos_ms % 60000) / 1000),
					(this->time_len_ms / 60000), ((this->time_len_ms % 60000) / 1000) );
	} else {
		snprintf(sout, sizeof(sout), "%d:%2.2d:%2.2d / %d:%2.2d:%2.2d", (new_timepos_ms / 3600000), (new_timepos_ms / 60000) % 60, ((new_timepos_ms % 60000) / 1000),
					(this->time_len_ms / 3600000), (this->time_len_ms / 60000) % 60, ((this->time_len_ms % 60000) / 1000) );
	}
	this->number->setText(QString::fromLatin1(sout));
}

void DispQtVideoControlbar::videoctrlbar_option_timemode_change(unsigned int mode)
{
	//emit this->seek_bar->signal_mainwinseekbar_option_timemode_change(mode);
}

void DispQtVideoControlbar::videoctrlbar_seekbar_send_seekpreview(void *preview_frame_p)
{
	if(!this->isHidden() && !this->timer_fadeout.isActive())
		this->seek_slider->seekslider_preview_send(preview_frame_p);
}
